//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: AllFood.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Drawing;

namespace AntGP
{
	/// <summary>
	/// This class is the collection of all food on the map.
    /// It also maintains counts of the amount of food detected and returned to the nest.
	/// </summary>
	public class AllFood : List<Food>
	{
        /// <summary>
        /// Default constructor
        /// </summary>
		public AllFood()
		{
		}

        /// <summary>
        /// Remove a piece of food from the collection based on the location of the food in the map.
        /// </summary>
        /// <remarks>
        /// Note: This is not the function used to pick up the food.
        /// The <see cref="Food.PickUpFood"/> function is used for that.
        /// This will completely remove the food from the map.
        /// </remarks>
        /// <param name="p">Location of the food.</param>
		public void Remove( Point p )
		{
            foreach (Food food in this)
            {
                if (food.Location.Equals(p))    // only one piece of food can be at a given location
                {
                    Remove(food);
                    return;
                }
            }
		}

        /// <summary>
        /// Reset the FoundThisTurn, MovedThisTurn, and LeadAntMovingThisFood variables
        /// for each piece of food in the collection.
        /// </summary>
		public void ResetFoodForNewTurn()
		{
			foreach( Food food in this )
			{
				food.FoundThisTurn = false;
				food.MovedThisTurn = false;
				food.LeadAntMovingThisFood = null;
			}
		}

        /// <summary>
        /// Has every piece of food been returned to the nest.
        /// </summary>
        /// <returns>True if all of the food has been returned.</returns>
		public bool IsAllFoodCollected()
		{
			foreach( Food food in this )
			{
				if (!food.HasBeenCollected)
				{
					return false;
				}
			}
			return true;
		}

        /// <summary>
        /// How many pieces of food have not been returned to the nest.
        /// </summary>
        /// <returns>The number of pieces of food that have not been returned to the nest.</returns>
        public int FoodNotCollected()
        {
            int cnt = 0;

            foreach (Food food in this)
            {
                if (!food.HasBeenCollected)
                {
                    cnt++;
                }
            }
            return cnt;
        }

        /// <summary>
        /// How many pieces of food have not been picked up.
        /// </summary>
        /// <returns>The number of pieces of food that have not been picked up.</returns>
        public int FoodNotLocated()
        {
            int cnt = 0;

            foreach (Food food in this)
            {
                if (!food.HasBeenFound)
                {
                    cnt++;
                }
            }
            return cnt;
        }
	}
}
