//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Program.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;

namespace AntGP
{
	/// <summary>
	/// This class represents one ant program.
    /// This class also contains functions to make random programs and evaluate programs.
	/// </summary>
	public class Program
	{
        /// <summary>
        /// The enum for the premade programs that can be created.
        /// </summary>
		public enum CannedPrograms
		{
            /// <summary>Hand writted solution that will solve problems with a map containing piles of food with weight 1.</summary>
			SimpleLightGroupedFoodSolution,
            /// <summary>Hand writted solution that will solve problems with a map containing food requiring more than one ant to lift.</summary>
            SimpleHeavyFoodSolution,
            /// <summary>Hand writted solution that will solve problems with a map that has water separating the nest and the food, using leaves to cross the river.</summary>
            TreeRiverSolution,
            /// <summary>Hand writted solution that will solve problems with a map that has water separating the nest and the food., using dead ant bridges to cross the river.</summary>
            DeadAntRiverSolution,
            /// <summary>Hand writted solution that will solve problems with a map that has water separating the nest and the food, and the food requires more than one ant to lift.</summary>
			HeavyFoodwithRiverSolution
		}

        /// <summary>
        /// Default Constructor
        /// </summary>
        public Program()
		{
		}

        /// <summary>
        /// Constructor that creates a program from a string of instructions.
        /// </summary>
        /// <remarks>
        /// The format of the program string should look like "if ( CarryingFood ) { ReleasePheromone; MoveToNest; } else { PickUpFood; }"
        /// All if and else blocks should be enclosed in { }. NoOp and TwoStatement tokens are not necessary. ; should come after each leaf token.
        /// </remarks>
        /// <param name="program">The program string.</param>
		public Program(string program)
		{
			LoadProgramString( program );
            CalculateSizeOfProgram();
		}

        /// <summary>
        /// Constructor for loading a canned program.
        /// </summary>
        /// <param name="cannedProgram">The prewritten program to load.</param>
		public Program( CannedPrograms cannedProgram )
		{
			LoadCannedProgram( cannedProgram );
            CalculateSizeOfProgram();
		}

        /// <summary>
        /// Constructor that creates a random program based on settings in the Genetic Program Options object.
        /// </summary>
        /// <param name="options">Options used to make the random program, including the allowable instruction set and the random number generator.</param>
        public Program(GeneticProgramOptions options)
        {
            ProgramTree = new ProgramNode();
            ProgramTree.Instruction = Tokens.GetRandomNonLeafNode(options);
            programSize = 1;
            CreateRandomChildrenNodes(options, ProgramTree);
        }

        /// <summary>
        /// Recursively copies a program tree from one program to another.
        /// This is used as part of crossover.
        /// </summary>
        /// <param name="myNode">The node in this programs tree.</param>
        /// <param name="originalNode">The node in the source program tree.</param>
        private void CopyProgramRecursive(ProgramNode myNode, ProgramNode originalNode)
        {
            myNode.Instruction = originalNode.Instruction;
            if (!Tokens.IsLeafNode(myNode.Instruction))
            {
                myNode.Left = new ProgramNode();
                CopyProgramRecursive(myNode.Left, originalNode.Left);

                myNode.Right = new ProgramNode();
                CopyProgramRecursive(myNode.Right, originalNode.Right);
            }
        }

        /// <summary>
        /// Recursively copies a program tree from one program to this one.
        /// This is used as part of crossover.
        /// </summary>
        /// <param name="program">The program tree to copy into this program.</param>
        public void CopyProgram(Program program)
        {
            ProgramTree = new ProgramNode();
            CopyProgramRecursive(ProgramTree, program.ProgramTree);
            CalculateSizeOfProgram();
        }

        /// <summary>
        /// Private for the binary program tree for this program object.
        /// </summary>
		private ProgramNode programTree;
        /// <summary>
        /// The binary program tree for this program object.
        /// </summary>
		public ProgramNode ProgramTree
		{
			set 
			{
				programTree = value;
			}
			get
			{
				return programTree;
			}
		}

        /// <summary>
        /// Private for the fitness for this program.
        /// </summary>
        private long fitness;
        /// <summary>
        /// The fitness for this program when it was last tested.
        /// Different random numbers values between executions can cause different values.
        /// </summary>
        public long Fitness
        {
            set
            {
                fitness = value;
            }
            get
            {
                return fitness;
            }
        }

        /// <summary>
        /// Private for the amount of food that was left when this program executed.
        /// </summary>
        private int foodLeft;
        /// <summary>
        /// The amount of food left (not returned to nest) the last time this program was executed.
        /// </summary>
        public int FoodLeft
        {
            set
            {
                foodLeft = value;
            }
            get
            {
                return foodLeft;
            }
        }

        /// <summary>
        /// Private for the amount of food that was not found when this program executed.
        /// </summary>
        private int foodNotFound;
        /// <summary>
        /// The amount of food not found (not picked up) the last time this program was executed.
        /// </summary>
        public int FoodNotFound
        {
            set
            {
                foodNotFound = value;
            }
            get
            {
                return foodNotFound;
            }
        }

        /// <summary>
        /// Private for the amount of water remaining on the map the last time this program was executed.
        /// </summary>
        private int waterRemaining;
        /// <summary>
        /// The amount of water remaining on the map the last time this program was executed.
        /// </summary>
        public int WaterRemaining
        {
            set
            {
                waterRemaining = value;
            }
            get
            {
                return waterRemaining;
            }
        }

        /// <summary>
        /// Private for the number of times this program was executed for each ant.
        /// </summary>
        private int numberOfTurns;
        /// <summary>
        /// The number of times this program was executed for each ant the last time this program was executed.
        /// If the ants didn't find all of the food, then the program would have executed the maximum number of times, as configured by the Genetic Program Options.
        /// </summary>
        public int NumberOfTurns
        {
            set
            {
                numberOfTurns = value;
            }
            get
            {
                return numberOfTurns;
            }
        }

        /// <summary>
        /// Recursive function that creates a random program.
        /// </summary>
        /// <param name="options">Genetic Program options that contains the allowable instruction set list, the random number generator, and the percent chance that any given node is a leaf statement.</param>
        /// <param name="node">The node to create child program nodes for.</param>
        private void CreateRandomChildrenNodes(GeneticProgramOptions options, ProgramNode node)
        {
            programSize += 2;

            node.Left = new ProgramNode();
            if ((options.MaximumProgramSize - 2 <= programSize) || (options.InitialPopulationRandom(100) < options.PercentLeafNode))
            {
                node.Left.Instruction = Tokens.GetRandomLeafNode(options);
            }
            else
            {
                node.Left.Instruction = Tokens.GetRandomNonLeafNode(options);
                CreateRandomChildrenNodes(options, node.Left);
            }

            node.Right = new ProgramNode();
            if ((options.MaximumProgramSize - 2 < programSize) || (options.InitialPopulationRandom(100) < options.PercentLeafNode))
            {
                node.Right.Instruction = Tokens.GetRandomLeafNode(options);
            }
            else
            {
                node.Right.Instruction = Tokens.GetRandomNonLeafNode(options);
                CreateRandomChildrenNodes(options, node.Right);
            }
        }

        /// <summary>
        /// When converting a program string into a program binary tree, this will find the end } that matches the bigging { in a program block.
        /// </summary>
        /// <param name="program">The program string to search.</param>
        /// <param name="beginBlock">The location in the program string of the {.</param>
        /// <returns>The location in the program string of the }. Returns -1 if it is not found.</returns>
        private int FindMatchingCurly(string program, int beginBlock)
        {
            int depth = 1;
            int endBlock = beginBlock+1;
            char[] beginEndChars = new char[2];
            beginEndChars[0] = '{';
            beginEndChars[1] = '}';

            while ( endBlock != -1 )
            {
                endBlock = program.IndexOfAny(beginEndChars, endBlock+1);
                if (endBlock == -1) // We didn't find the closing }. This is a bad program block.
                {
                    return endBlock;
                }
                if (program.Substring(endBlock, 1).Equals("{")) // We found another opening {, so that is one more closing } we have to find.
                {
                    depth++;
                }
                else // We found a closing }.
                {
                    depth--;
                    if (depth == 0) // The closing } is the one that matches our opening {. 
                    {
                        return endBlock;
                    }
                }
            }

            return -1;
        }

        /// <summary>
        /// Recursively creates the program binary tree from a string.
        /// </summary>
        /// <param name="node">The node that is being created in this recursive call.</param>
        /// <param name="program">The part of the program string that goes with this node.</param>
        private void LoadProgramRecursive(ProgramNode node, string program)
		{
			string instruction = null;
            int beginBlock, endBlock;

			if ( program.Substring( 0, 2 ).Equals( "if" ))
			{
				beginBlock = program.IndexOf("{");
                if (beginBlock == -1)
                {
                    throw new Exception("Missing open block on if");
                }
                
                instruction = program.Substring(0, beginBlock);
                node.Instruction = Tokens.StringToInstruction( instruction );

                endBlock = FindMatchingCurly(program, beginBlock);
                if (endBlock == -1)
                {
                    throw new Exception("Missing close block on if " + instruction);
                }

                node.Left = new ProgramNode();
                LoadProgramRecursive( node.Left, program.Substring( beginBlock+1, endBlock-beginBlock-1 ));

                beginBlock = program.IndexOf("{", endBlock);
                if (beginBlock == -1)
                {
                    throw new Exception("Missing open block on else " + instruction);
                }
                endBlock = FindMatchingCurly(program, beginBlock);
                if (endBlock == -1)
                {
                    throw new Exception("Missing close block on else " + instruction);
                }
				
				node.Right = new ProgramNode();
                LoadProgramRecursive( node.Right, program.Substring( beginBlock+1, endBlock-beginBlock-1 ));
			}
			else
			{
                beginBlock = program.IndexOf(";");
                if (beginBlock == -1)
                {
                    throw new Exception("No ; after instruction");
                }

                instruction = program.Substring(0, beginBlock);

                if ( beginBlock + 1 == program.Length)
                {
                    // if only one instruction in this block, put it in this node.
                    node.Instruction = Tokens.StringToInstruction( instruction );
                }
                else
                {
                    // if multiple instructions in this block, put a TwoStmt (split) in this node,
                    // and put this instruction in the left node.
                    // The right node gets everything else.
                    node.Instruction = Tokens.Instruction.TwoStatements;
                    node.Left = new ProgramNode();
                    node.Left.Instruction = Tokens.StringToInstruction( instruction );

                    //	/\      If this block has multiple statemets, the will get set up like this.
			        //   /\
			        //    /\
                    node.Right = new ProgramNode();
                    LoadProgramRecursive( node.Right, program.Substring( beginBlock+1 ));
                }
			}
		}

        /// <summary>
        /// Loads the program string and turns it into a binary program tree.
        /// </summary>
        /// <param name="program">The program string to load.</param>
		private void LoadProgramString( string program )
		{
            // Remove all white spaces and parens. By doing this, if ( FoodHere ) turns into ifFoodHere, which is the token string.
			program = program.Replace( " ", "" );
			program = program.Replace( "(", "" );
			program = program.Replace( ")", "" );

            ProgramTree = new ProgramNode();
            LoadProgramRecursive(ProgramTree, program);
		}

        /// <summary>
        /// Loads a premade canned program into a binary program tree.
        /// </summary>
        /// <param name="cannedProgram">The canned program to load.</param>
		private void LoadCannedProgram( CannedPrograms cannedProgram )
		{
			string program = "NOOP";
			switch (cannedProgram)
			{
				case CannedPrograms.SimpleLightGroupedFoodSolution:
                    program = "if ( CarryingFood ) { ReleasePheromone; MoveToNest; } else { if ( FoodHere ) { PickUpFood; } else { if ( MoveToAdjacentFood ) { PickUpFood; } else { if ( MoveToPheromoneAwayFromNest ) { PickUpFood; } else { MoveQuasiRandom; } } } }";
                    LoadProgramString(program);
					break;
				case CannedPrograms.SimpleHeavyFoodSolution:
                    program = "if ( CantLiftFood ) { if ( TiredOfWaiting ) { EscapePheromone; } else { ReleasePheromone; } } else { if ( CarryingFood ) { ReleasePheromone; MoveToNest; } else { if ( FoodHere ) { PickUpFood; } else { if ( MoveToAdjacentFood ) { PickUpFood; } else { if ( MoveToPheromoneAwayFromNest ) { PickUpFood; } else { MoveQuasiRandom; } } } } }";
                    LoadProgramString(program);
					break;
                case CannedPrograms.TreeRiverSolution:
                    program = "if ( CarryingFood ) { MoveToNest; MoveQuasiRandom; } else { if ( MoveToAdjacentFood ) { if ( AnotherAntHere ) { PickUpFood; } else { if ( MoveToAdjacentFood ) { MoveRandom; ReleasePheromone; } else { PickUpFood; } } } else { if ( FoodHere ) { PickUpFood; } else { if ( CarryingLeaf ) { if ( WaterAhead ) { DropLeafInWater; } else { MoveQuasiRandom; } } else { if ( MoveToAdjacentLeaf ) { PickUpLeaf; } else { if ( MoveToStrongestPheromone ) { MoveQuasiRandom; } else { MoveQuasiRandom; } } } } } }";
                    LoadProgramString(program);
                    break;
				case CannedPrograms.DeadAntRiverSolution:
                    program = "if ( CarryingFood ) { ReleasePheromone; MoveToNestThroughWater; } else { if ( MoveToAdjacentFood ) { if ( AnotherAntHere ) { PickUpFood; } else { if ( MoveToAdjacentFood ) { MoveRandom; ReleasePheromone; } else { PickUpFood; } } } else { if ( FoodHere ) { PickUpFood; } else { if ( MoveToDeadAnt ) { ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; MoveForward; MoveForward; MoveForward; MoveIntoWater; } else { if ( MoveToPheromoneAwayFromNest ) { MoveForward; if ( MoveIntoWater ) { ReleasePheromone; ReleasePheromone; } else { MoveForward; } } else { if ( PheromoneHere ) { MoveRandom; } else { if ( WaterAhead ) { ReleasePheromone; ReleasePheromone; ReleasePheromone; MoveIntoWater; } else { MoveQuasiRandom; } } } } } } }";
                    LoadProgramString(program);
					break;
				case CannedPrograms.HeavyFoodwithRiverSolution:
                    program = "if ( CantLiftFood ) { if ( TiredOfWaiting ) { DropFood; EscapePheromone; } else { ReleasePheromone; } } else { if ( CarryingFood ) { MoveToNestThroughWater; } else { if ( MoveToAdjacentFood ) { if ( AnotherAntHere ) { PickUpFood; } else { if ( MoveToAdjacentFood ) { MoveRandom; ReleasePheromone; } else { PickUpFood; } } } else { if ( FoodHere ) { PickUpFood; } else { if ( MoveToDeadAnt ) { ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; ReleasePheromone; if ( WaterAhead ) { MoveIntoWater; } else { EscapePheromone; } } else { if ( MoveToStrongestPheromone ) { if ( FoodHere ) { PickUpFood; } else { if ( ObstacleAhead ) { EscapePheromone; } else { PickUpFood; } } if ( MoveIntoWater ) { ReleasePheromone; ReleasePheromone; } else { MoveForward; } } else { if ( PheromoneHere ) { MoveRandom; } else { if ( WaterAhead ) { ReleasePheromone; ReleasePheromone; ReleasePheromone; MoveIntoWater; } else { MoveQuasiRandom; } } } } } } } }";
                    LoadProgramString(program);
					break;
			}

			LoadProgramString( program );
		}

        /// <summary>
        /// Determines the fitness of a program after it has been executed. A lower fitness is better.
        /// </summary>
        /// <param name="options">The Genetic Program Options indicate the amount that each parameter will impact the fitness.</param>
        /// <param name="foodLeft">The amount of food that was not returned to the nest.</param>
        /// <param name="foodNotFound">The amount of food that was not picked up.</param>
        /// <param name="waterRemaining">The amount of water that was not changed to a dead ant or leaf on water.</param>
        /// <param name="numberOfTurns">The numeber of turns it took to find all the food, or the maximum number of turns allowed by the interpreter.</param>
        /// <param name="numberOfNodes">The number of nodes in the program tree.</param>
        /// <returns>The programs fitness.</returns>
        public static long DetermineFitness(GeneticProgramOptions options, int foodLeft, int foodNotFound, int waterRemaining, int numberOfTurns, int numberOfNodes )
        {
            return (long)foodLeft * (long)options.FitnessRemainingFood +
                    (long)foodNotFound * (long)options.FitnessUndiscoveredFood +
                    (long)waterRemaining * (long)options.FitnessWaterRemaining +
                    ((long)((long)numberOfTurns * (long)options.FitnessNumberOfTurns) / 100L) +
                    (long)(numberOfNodes / (long)options.FitenssProgramComplexityGrouping) * (long)options.FitnessProgramComplexity;
        }


        /// <summary>
        /// Private variable for the size of the program tree.
        /// </summary>
        private int programSize = 0;
        public int ProgramSize
        {
            get
            {
                return programSize;
            }
        }

        /// <summary>
        /// Recursively determines the size of the program tree.
        /// </summary>
        /// <param name="node">The node to get the size of.</param>
        private void CalculateSizeOfProgramRecursive(ProgramNode node)
        {
            if (node != null)
            {
                programSize++;
                CalculateSizeOfProgramRecursive(node.Left);
                CalculateSizeOfProgramRecursive(node.Right);
            }
        }

        /// <summary>
        /// Gets the number of program nodes in the program tree.
        /// </summary>
        /// <returns>The number of program nodes in the program tree.</returns>
        public int CalculateSizeOfProgram()
        {
            programSize = 0;
            CalculateSizeOfProgramRecursive(ProgramTree);
            return programSize;
        }
        
        /// <summary>
        /// Determines the fitness of a program by running it through the interpreter.
        /// </summary>
        /// <param name="options">Genetic Program options used by the interpreter.</param>
        public void EvaluateFitness(GeneticProgramOptions options)
        {
            Interpreter interpreter = new Interpreter();
            interpreter.Options = options;
            interpreter.Program = this;
            interpreter.ExecuteProgram(null);

            Fitness = DetermineFitness(options, FoodLeft, FoodNotFound, WaterRemaining, NumberOfTurns, ProgramSize);
        }
	}
}
