//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: TurnCompleteEventArgs.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;

namespace AntGP
{
	/// <summary>
	/// Event arguments used every time the Ant program interpreter has executed the program on each ant once.
	/// </summary>
	public class TurnCompleteEventArgs
	{
        /// <summary>
        /// Default Constructor.
        /// </summary>
		public TurnCompleteEventArgs()
		{
		}

        /// <summary>
        /// Constructor with Event Arguments.
        /// </summary>
        /// <param name="map">The map that the ants are moving around in.</param>
        /// <param name="ants">The ants, their current positions, and if they are carrying food.</param>
        /// <param name="numberOfNodes">The number of nodes in the ant program binary tree.</param>
        /// <param name="turn">The number of times each ant has executed the ant program so far.</param>
        /// <param name="fitnessTurns">The total number of times all ants have moved divided by the number of ants.
        /// This is a true count of the number of turns, because it doesn't reward large programs that simply have twice as many move statements to finish the problem faster.</param>
        /// <param name="complete">Is the Ant Program complete.</param>
        public TurnCompleteEventArgs(Map map, Colony ants, int numberOfNodes, int turn, int averageAntMovement, bool complete)
		{
			this.map = map;
            this.ants = ants;
            this.numberOfNodes = numberOfNodes;
            this.turn = turn;
            this.averageAntMovement = averageAntMovement;
            this.complete = complete;
		}

        /// <summary>
        /// Private of the map that the ants are moving around on.
        /// </summary>
		private Map map;
        /// <summary>
        /// The map that the ants are moving around on.
        /// </summary>
		public Map Map
		{
			get 
			{
				return map;
			}
		}
        /// <summary>
        /// Private for the list of ants that have been executing the ant program.
        /// </summary>
        private Colony ants;
        /// <summary>
        /// The list of ants that have been executing the ant program.
        /// </summary>
        public Colony Ants
        {
            get
            {
                return ants;
            }
        }
        /// <summary>
        /// Private for the number of program nodes in the ant program.
        /// </summary>
        private int numberOfNodes;
        /// <summary>
        /// The number of program nodes in the ant program.
        /// </summary>
        public int NumberOfNodes
        {
            get
            {
                return numberOfNodes;
            }
        }

        /// <summary>
        /// Private for the number of times each ant has executed the ant program so far
        /// </summary>
        private int turn;
        /// <summary>
        /// The number of times each ant has executed the ant program so far
        /// </summary>
        public int Turn
        {
            get
            {
                return turn;
            }
        }

        /// <summary>
        /// Private for the total number of times all ants have moved divided by the number of ants.
        /// </summary>
        private int averageAntMovement;
        /// <summary>
        /// The total number of times all ants have moved divided by the number of ants.
        /// This is a true count of the number of turns, because it doesn't reward large programs that simply have twice as many move statements to finish the problem faster.
        /// </summary>
        public int AverageAntMovement
        {
            get
            {
                return averageAntMovement;
            }
        }

        /// <summary>
        /// Private indicator that the ant program is complete.
        /// </summary>
        private bool complete = false;
        /// <summary>
        /// Is the ant program complete. True if the user stopped the program or if all the food was found.
        /// </summary>
        public bool Complete
        {
            get
            {
                return complete;
            }
        }

        /// <summary>
        /// The amount of food that has not been returned to the nest so far.
        /// </summary>
        public int FoodNotCollected
        {
            get
            {
                return map.FoodNotCollected();
            }
        }

        /// <summary>
        /// The amount of food that has never been picked up so far.
        /// </summary>
        public int FoodNotLocated
        {
            get
            {
                return map.FoodNotLocated();
            }
        }

        /// <summary>
        /// The amount of water that has not been changed to a dead ant or a leaf on water.
        /// </summary>
        public int WaterRemaining
        {
            get
            {
                return map.WaterRemaining();
            }
        }

        /// <summary>
        /// Private for the speed that the ant program will execute.
        /// </summary>
        private int refreshSpeed = 0;
        /// <summary>
        /// The speed that the ant program will execute. This is set by the event listener, and causes the ant program to sleep to slow down the simulator GUI.
        /// </summary>
        public int RefreshSpeed
        {
            set
            {
                refreshSpeed = value;
            }
            get
            {
                return refreshSpeed;
            }
        }
	}
}
